# -*- coding: utf-8 -*-
from datetime import datetime
from khayyam import JalaliDatetime
from tg import request
__author__ = 'vahid'

class Formatter(object):
    def __init__(self, value):
        self.value = value

    def format_datetime(self):
        if request and hasattr(request, 'language') and request.language.startswith('fa'):
            dt = JalaliDatetime.from_datetime(self.value)
        else:
            dt = self.value
        return dt.strftime('%Y-%m-%d %H:%M:%S')

    def __unicode__(self):
        if not self.value:
            return ''
        if isinstance(self.value, datetime):
            value = self.format_datetime()
        else:
            value = self.value
        return value


# -*- coding: utf-8 -*-
__author__ = 'vahid'

class Field(object):
    def __init__(self, sa_column, **kw):
        self.sa_column = sa_column
        for k, v in kw.items():
            setattr(self, k, v)

    @property
    def name(self):
        return self.sa_column.name

    @property
    def label(self):
        if hasattr(self, 'alias'):
            return self.alias
        return self.name

    @property
    def display_label(self):
        title = self.label.replace('_', ' ')
        return title[:1].upper() + title[1:]

    @property
    def type(self):
        return self.sa_column.type



# -*- coding: utf-8 -*-
# import sqlalchemy
# from sqlalchemy.sql.elements import Label as sql_label
from tg import request
from tg.render import render
from sqlalchemy import desc
from collections import OrderedDict
from markupsafe import Markup
from .sort import Sort
from .format import Formatter
from .column import Column as TableColumn
from .sa_helpers import extract_columns
__author__ = 'vahid'

class SmartTable(object):
    template = "tgtoolbox.smart_crud.templates.smart_table"
    allow_new = True
    allow_delete = True
    allow_edit = True
    limit_max_rows = None
    limit_min_rows = None

    def __init__(self, session, query=None, paginator_key='data'):
        self.session = session # Not used yet !
        self._query = query

        self.total_count = 0
        self.can_new = False
        self.can_delete = False
        self.can_edit = False

        self.sorts = []
        self.columns = []
        self.paginator_key = paginator_key

    def get_query(self):
        return self._query
    def set_query(self, value):
        self._query = value
        self.total_count = value.count()
        self.can_new = self.allow_new and (not self.limit_max_rows or self.total_count < self.limit_max_rows)
        self.can_delete = self.allow_delete and (not self.limit_min_rows or self.total_count > self.limit_min_rows)
        self.can_edit = self.allow_edit
        self.columns = self._extract_columns()
        self.sorts = Sort.from_request()
    query = property(get_query, set_query)


    def _extract_columns(self):
        columns = OrderedDict()
        omit_fields = [] if not hasattr(self, '__omit_fields__') else self.__omit_fields__
        for alias, sa_column in extract_columns(self._query):
            if alias in omit_fields:
                continue
            columns[alias] = TableColumn(sa_column, table=self, alias=alias)
        return columns

    def format(self, row, column):
        value = getattr(row, column.label)
        if hasattr(self, '__xml_fields__') and column.label in self.__xml_fields__:
            if hasattr(self, column.label):
                value = getattr(self, column.label)(row)
            return Markup(value)
        return Formatter(value).__unicode__()

    @property
    def paginator(self):
        return request.paginators[self.paginator_key]

    @property
    def page_size(self):
        return self.paginator.paginate_items_per_page

    @property
    def page_index(self):
        return self.paginator.paginate_page - 1

    def paginate(self):
        start = self.page_size * self.page_index
        end = start + self.page_size
        self._query = self._query[start:end]

    def sort(self):
        create_sort_expr = lambda c, s: c.sa_column if not s.desc else desc(c.sa_column)
        expressions = []
        for sort in self.sorts:
            expressions.append(create_sort_expr(self.columns[sort.name], sort))
        self._query = self._query.order_by(*expressions)

    def display(self):
        self.sort()
        self.paginate()
        return render(dict(table=self),
                      template_name=self.template)

    def is_sortable(self, col):
        # TODO: sortable columns
        return True



# -*- coding: utf-8 -*-
from markupsafe import Markup
from tg import request, url
from .field import Field
__author__ = 'vahid'


class Column(Field):

    def render_header(self):
        sorts = self.table.sorts.clone()
        sorts.add_or_toggle(self.label)

        url_params = dict(request.GET)
        url_params['sort_cols'] = unicode(sorts)
        url_ = url(request.path, url_params)

        sorts.remove(self.label)
        url_params['sort_cols'] = unicode(sorts)
        remove_url = url(request.path, url_params)

        if self.table.is_sortable(self):
            s = self.table.sorts.get(self.label)
            if s:
                icon = '<a href="%s"><i class="glyphicon glyphicon-remove"/></i></a>&nbsp;' % remove_url
                icon += '<a href="%(url)s"><i class="glyphicon glyphicon-arrow-%(icon_name)s"/></i></a>' % dict(
                    url=url_,
                    icon_name="down" if s.desc else "up"
                )
            else:
                icon = ''

            result = '%(icon)s<a href="%(url)s">%(title)s</a>' % \
                   dict(title=self.display_label,
                        icon=icon,
                        url=url_)
        else:
            result = self.display_label

        return Markup(result)


# -*- coding: utf-8 -*-
from .controller import SmartCrudRestController
from .column import Column as TableColumn
from .table import SmartTable
from .pagination import SmartPaginate as paginate
from .form import SmartAddForm
from .sort import Sort, SortItem
from .format import Formatter
__author__ = 'vahid'



__author__ = 'vahid'

from tgext.crud import EasyCrudRestController

def just_editable_actions_maker(table_filler, obj):
    primary_fields = table_filler.__provider__.get_primary_fields(table_filler.__entity__)
    pklist = '/'.join(map(lambda x: str(getattr(obj, x)), primary_fields))
    #value = '<div><div>&nbsp;<a href="'+pklist+'/edit" style="text-decoration:none">edit</a></div></div>'
    return '''
    <a href="%s/edit" class="btn btn-primary">
        <span class="glyphicon glyphicon-pencil"></span>
    </a>
    ''' % pklist


class JustEditableCrudRestController(EasyCrudRestController):

    def post(self, *args, **kw):
        """This is not allowed."""
        pass

    def post_delete(self, *args, **kw):
        """This is not allowed."""
        pass



# -*- coding: utf-8 -*-
from tg import request
from collections import OrderedDict
__author__ = 'vahid'

class SortItem(object):
    def __init__(self, name, descending=False):
        self.name = name
        self.desc = descending

    @classmethod
    def parse(cls, str):
        return cls(str[1:], descending=str[0] == '-')

    def clone(self):
        return self.__class__(self.name, self.desc)

    def __unicode__(self):
        return U'%s%s' % (u'-' if self.desc else u'+', self.name)

class Sort(object):

    def __init__(self, sort_items=None):
        self.items = OrderedDict()
        if sort_items:
            for item in sort_items:
                self.items[item.name] = item

    def add(self, col_name, descending=False):
        self.items[col_name] = SortItem(col_name, descending)

    def add_or_toggle(self, col_name):
        if col_name in self.items:
            self.items[col_name].desc ^= True
        else:
            self.add(col_name)

    def __iter__(self):
        return iter(self.items.values())

    @classmethod
    def parse(cls, str):
        if not str:
            return cls()
        return cls([SortItem.parse(i) for i in str.split('|')])

    @classmethod
    def from_request(cls):
        return cls.parse(request.GET.get('sort_cols'))

    def __unicode__(self):
        return u'|'.join([unicode(i) for i in self.items.values()])

    def __len__(self):
        return len(self.items)

    def clone(self):
        return self.__class__([i.clone() for i in self.items.values()])

    def get(self, key, default=None):
        return self.items.get(key, default)

    def remove(self, col_name):
        del self.items[col_name]


# -*- coding: utf-8 -*-
from tw2.recaptcha import ReCaptchaWidget as BadReCaptchaWidget
from tw2.recaptcha.validator import ReCaptchaValidator as BadReCaptchaValidator
__author__ = 'vahid'

class ReCaptchaWidget(BadReCaptchaWidget):
    """
    Hack by Vahid Mardani, due the tw2.recaptcha validation bug
    """
    _sub_compound = True
    pass

class ReCaptchaValidator(BadReCaptchaValidator):
    """
    Hack by Vahid Mardani, due the tw2.recaptcha validation bug
    """
    not_empty = True


# -*- coding: utf-8 -*-
import tw2.core as twc
__author__ = 'vahid'


class ImageDisplay(twc.Widget):
    value = twc.Param(default='/img/no-image-available.jpg', attribute=True, view_name='src')
    width = twc.Param(default=None, attribute=True)
    height = twc.Param(default=None, attribute=True)
    alt = twc.Param(default=None, attribute=True)
    template = 'mako:tgtoolbox.widgets.templates.image_display'


# -*- coding: utf-8 -*-
__author__ = 'vahid'

import os.path
from sqlalchemy import event
from sqlalchemy.orm.attributes import InstrumentedAttribute
from tgext.datahelpers.fields import Attachment as DataHelperAttachment
from tgext.datahelpers.attachments import AttachedImage as DataHelperAttachedImage, AttachedFile

try:
    from PIL import Image
except ImportError:
    import Image


class Attachment(DataHelperAttachment):

    def process_bind_param(self, value, dialect):
        if isinstance(value, NullAttachedImage):
            value = None
        return super(Attachment, self).process_bind_param(value, dialect)

    def process_result_value(self, value, dialect):
        result = super(Attachment, self).process_result_value(value, dialect)
        if result:
            return result
        return self.attachment_type.create_null_attachment()


class NullAttachedImage(object):
    #url = '/img/no-image-available.jpg'
    thumb_url = url ='/img/no-image-available.jpg'

    def __eq__(self, other):
        return other is None

    def __ne__(self, other):
        return other is not None

    def __nonzero__(self):
        return False


class NullGrayscaleAttachedImage(NullAttachedImage):
    gray_thumb_url = NullAttachedImage.url


class AttachedImage(DataHelperAttachedImage):

    def write(self):
        AttachedFile.write(self)

        if getattr(self.file, 'name', None) != self.local_path:
            self.file.seek(0)
            thumbnail = Image.open(self.file)
            thumbnail.thumbnail(self.thumbnail_size, Image.ANTIALIAS)
            thumbnail = thumbnail.convert('RGBA')
            thumbnail.format = self.thumbnail_format
            thumbnail.save(self.thumb_local_path)

    @classmethod
    def create_null_attachment(cls):
        return NullAttachedImage()


class GrayscaleAttachedImage(AttachedImage):

    def __init__(self, file, filename, uuid=None):
        super(GrayscaleAttachedImage, self).__init__(file, filename, uuid=uuid)

        gray_thumb_filename = 'thumb-gray.'+self.thumbnail_format.lower()
        self.gray_thumb_local_path = os.path.join(self.attachment_dir, gray_thumb_filename)
        self.gray_thumb_url = '/'.join([self.attachments_url, self.uuid, gray_thumb_filename])

    def write(self):
        super(GrayscaleAttachedImage, self).write()

        if getattr(self.file, 'name', None) != self.local_path:
            self.file.seek(0)
            thumbnail = Image.open(self.file)
            thumbnail.thumbnail(self.thumbnail_size, Image.ANTIALIAS)
            thumbnail = thumbnail.convert('LA') # Make grayscale
            thumbnail.format = self.thumbnail_format
            thumbnail.save(self.gray_thumb_local_path)

    @classmethod
    def create_null_attachment(cls):
        return NullGrayscaleAttachedImage()


class AttachmentCleanerMixin(object):

    @staticmethod
    def after_delete_banner(mapper, connection, target):
        for k, v in target.__table__.columns.items():
            if isinstance(v.type, Attachment):
                column_value = getattr(target, k)
                if column_value is not None and isinstance(column_value, AttachedFile):
                    column_value.unlink()

    @staticmethod
    def on_set(target, value, oldvalue, initiator):
        if oldvalue and isinstance(oldvalue, AttachedFile):
            oldvalue.unlink()

    @classmethod
    def register_events(cls):
        event.listen(cls, 'after_delete', cls.after_delete_banner)
        columns = {name: attr for name, attr in cls.__dict__.copy().iteritems()\
                   if isinstance(attr, InstrumentedAttribute)\
                        and hasattr(attr.property, 'columns')\
                        and len(attr.property.columns)\
                        and isinstance(attr.property.columns[0].type, Attachment)}
        for name, attr in columns.iteritems():
            event.listen(getattr(cls, name), 'set', cls.on_set)




# -*- coding: utf-8 -*-
import tw2.bootstrap.forms as twf
from tg import url
from sprox.formbase import AddRecordForm, EditableForm
from sprox.sa.widgetselector import SAWidgetSelector

__author__ = 'vahid'

class SmartWidgetSelector(SAWidgetSelector):
    def select(self, field):
        w = super(SmartWidgetSelector, self).select(field)
        w.css_class = 'form-control'
        return w


class SmartAddForm(AddRecordForm):
    __base_widget_type__= twf.BootstrapForm
    __widget_selector_type__ = SmartWidgetSelector
    __base_widget_args__ = {'action': url('./')}

class SmartEditForm(EditableForm):
    __base_widget_type__= twf.BootstrapForm
    __widget_selector_type__ = SmartWidgetSelector
    __base_widget_args__ = {'action': url('./')}




# -*- coding: utf-8 -*-
from tg import request, redirect, expose, tmpl_context, response, abort
from tg.decorators import without_trailing_slash, before_validate, with_trailing_slash
from tgext.crud.controller import CrudRestController, errors
from tgext.crud.utils import create_setter, map_args_to_pks, allow_json_parameters, force_response_type
from tgext.crud.decorators import registered_validate, catch_errors
from sprox.fillerbase import RecordFiller, AddFormFiller
from .table import SmartTable
from .form import SmartAddForm, SmartEditForm
from .pagination import SmartPaginate as paginate
__author__ = 'vahid'


class SmartCrudRestController(CrudRestController):
    model = None
    pagination = False
    menu_items = []
    allow_new = True
    allow_delete = True
    allow_edit = True
    limit_max_rows = None
    limit_min_rows = None

    def __init__(self, session, menu_items=None):
        self.session = session
        if menu_items:
            self.menu_items = menu_items

        if not (hasattr(self, 'table') or hasattr(self, 'table_type')):
            class Table(SmartTable):
                __entity__ = self.model
                allow_new = self.allow_new
                allow_delete = self.allow_delete
                allow_edit = self.allow_edit
                limit_max_rows = self.limit_max_rows
                limit_min_rows = self.limit_min_rows

            self.table = Table(session)

        if self.allow_edit:
            if not (hasattr(self, 'edit_form') or hasattr(self, 'edit_form_type')):
                class EditForm(SmartEditForm):
                    __entity__ = self.model
                self.edit_form = EditForm(session)

            if not (hasattr(self, 'edit_filler') or hasattr(self, 'edit_filler_type')):
                class EditFiller(RecordFiller):
                    __entity__ = self.model
                self.edit_filler = EditFiller(session)
        else:
            self.edit_form = None
            self.edit_filler = None

        if self.allow_new:
            if not (hasattr(self, 'new_form') or hasattr(self, 'new_form_type')):
                class NewForm(SmartAddForm):
                    __entity__ = self.model
                self.new_form = NewForm(session)

            if not (hasattr(self, 'new_filler') or hasattr(self, 'new_filler_type')):
                class NewFiller(AddFormFiller):
                    __entity__ = self.model
                self.new_filler = NewFiller(session)
        else:
            self.new_form = None
            self.new_filler = None

        super(SmartCrudRestController, self).__init__(session, menu_items)

        #Permit to quickly customize form options
        if hasattr(self, '__form_options__'):
            for name, value in self.__form_options__.items():
                for form in (self.edit_form, self.new_form):
                    if form:
                        setattr(form, name, value)

        #Permit to quickly create custom actions to set values
        if hasattr(self, '__setters__'):
            for name, config in self.__setters__.items():
                setattr(self, name, create_setter(self, self.get_all, config))


        #Permit to quickly customize table options
        if hasattr(self, '__table_options__'):
            for name, value in self.__table_options__.items():
                setattr(self.table, name, value)

    def _before(self, *args, **kw):
        tmpl_context.title = self.title
        tmpl_context.menu_items = self.menu_items
        tmpl_context.kept_params = self._kept_params()
        tmpl_context.crud_helpers = self.helpers
        tmpl_context.model_name = self.model.__name__

        for resource in self.resources:
            resource.inject()
        force_response_type(self.response_type)

    def _adapt_menu_items(self, menu_items):
        if hasattr(self, 'menu_items'):
            return self.menu_items
        return menu_items

    def select_query(self):
        return self.session.query(self.model)

    @with_trailing_slash
    @expose('tgtoolbox.smart_crud.templates.list')
    @paginate(items_per_page=20)
    def get_all(self, *args, **kw):
        self.table.query = self.select_query()
        tmpl_context.table = self.table
        return dict(data=self.table.query)

    @expose('mako:tgtoolbox.smart_crud.templates.view')
    @expose('json:')
    def get_one(self, *args, **kw):
        """get one record, returns HTML or json"""
        #this would probably only be realized as a json stream
        kw = map_args_to_pks(args, {})

        if request.response_type == 'application/json':
            obj = self.provider.get_obj(self.model, kw)
            if obj is None:
                response.status_code = 404
            elif self.conditional_update_field is not None:
                response.last_modified = getattr(obj, self.conditional_update_field)

            return dict(model=self.model.__name__,
                        value=self._dictify(obj))

        tmpl_context.widget = self.edit_form
        value = self.edit_filler.get_value(kw)
        return dict(value=value, model=self.model.__name__)

    @expose('mako:tgtoolbox.smart_crud.templates.edit')
    def edit(self, *args, **kw):
        """Display a page to edit the record."""
        if not self.allow_edit:
            abort(404)
        pks = self.provider.get_primary_fields(self.model)
        kw = map_args_to_pks(args, {})

        tmpl_context.widget = self.edit_form
        value = self.edit_filler.get_value(kw)
        value['_method'] = 'PUT'
        return dict(value=value, model=self.model.__name__, pk_count=len(pks))

    @without_trailing_slash
    @expose('mako:tgtoolbox.smart_crud.templates.new')
    def new(self, *args, **kw):
        """Display a page to show a new record."""
        if not self.allow_new:
            abort(404)
        tmpl_context.widget = self.new_form
        return dict(value=kw, model=self.model.__name__)

    @expose(content_type='text/html')
    @expose('json:', content_type='application/json')
    @before_validate(allow_json_parameters)
    @catch_errors(errors, error_handler=new)
    @registered_validate(error_handler=new)
    def post(self, *args, **kw):
        if not self.allow_new:
            abort(404)
        obj = self.provider.create(self.model, params=kw)
        if request.response_type == 'application/json':
            if obj is not None and self.conditional_update_field is not None:
                response.last_modified = getattr(obj, self.conditional_update_field)

            return dict(model=self.model.__name__,
                        value=self._dictify(obj))

        return redirect('./', params=self._kept_params())

    @expose(content_type='text/html')
    @expose('json:', content_type='application/json')
    @before_validate(allow_json_parameters)
    @before_validate(map_args_to_pks)
    @registered_validate(error_handler=edit)
    @catch_errors(errors, error_handler=edit)
    def put(self, *args, **kw):
        """update"""
        if not self.allow_edit:
            abort(404)
        omit_fields = []
        if getattr(self, 'edit_form', None):
            omit_fields.extend(self.edit_form.__omit_fields__)

        for remembered_value in self.remember_values:
            value = kw.get(remembered_value)
            if value is None or value == '':
                omit_fields.append(remembered_value)

        obj = self.provider.get_obj(self.model, kw)

        #This should actually by done by provider.update to make it atomic
        can_modify = True
        if obj is not None and self.conditional_update_field is not None and \
           request.if_unmodified_since is not None and \
           request.if_unmodified_since < getattr(obj, self.conditional_update_field):
                can_modify = False

        if obj is not None and can_modify:
            obj = self.provider.update(self.model, params=kw, omit_fields=omit_fields)

        if request.response_type == 'application/json':
            if obj is None:
                response.status_code = 404
            elif can_modify is False:
                response.status_code = 412
            elif self.conditional_update_field is not None:
                response.last_modified = getattr(obj, self.conditional_update_field)

            return dict(model=self.model.__name__,
                        value=self._dictify(obj))

        pks = self.provider.get_primary_fields(self.model)
        return redirect('../' * len(pks), params=self._kept_params())

    @expose(content_type='text/html')
    @expose('json:', content_type='application/json')
    def post_delete(self, *args, **kw):
        """This is the code that actually deletes the record"""
        if not self.allow_delete:
            abort(404)
        kw = map_args_to_pks(args, {})

        obj = None
        if kw:
            obj = self.provider.get_obj(self.model, kw)

        if obj is not None:
            self.provider.delete(self.model, kw)

        if request.response_type == 'application/json':
            return dict()

        pks = self.provider.get_primary_fields(self.model)
        return redirect('./' + '../' * (len(pks) - 1), params=self._kept_params())

    @expose('mako:tgtoolbox.smart_crud.templates.delete_confirm')
    def get_delete(self, *args, **kw):
        """This is the code that creates a confirm_delete page"""
        if not self.allow_delete:
            abort(404)
        return dict(args=args)


# -*- coding: utf-8 -*-
from sqlalchemy import Column as SaColumn
from sqlalchemy.sql.elements import Label as SaLabel
from sqlalchemy.ext.declarative.api import DeclarativeMeta
__author__ = 'vahid'


def extract_columns(obj):
    if isinstance(obj, DeclarativeMeta):
        for c in obj.__table__.c.values():
            yield c.name, c
    else:
        for descr in obj.column_descriptions:
            expr = descr['expr']
            if isinstance(expr, SaLabel):
                yield expr.key, expr.element
            elif issubclass(descr['type'].__class__, DeclarativeMeta):
                for c in expr.__table__.c:
                    if isinstance(c, SaColumn):
                        yield c.name, c
            elif expr.is_attribute:
                yield expr.key, getattr(expr.class_, expr.key)



# -*- coding: utf-8 -*-
import tw2.core as twc
from sprox.widgets import \
    PasswordField as _PasswordField, \
    TextField as _TextField
__author__ = 'vahid'

__all__ = [
    'PasswordField',
    'TextField',
]

class PasswordField(_PasswordField):
    css_class = 'form-control'

class TextField(_TextField):
    css_class = 'form-control'





# -*- coding: utf-8 -*-
from tg.decorators import paginate
__author__ = 'vahid'

class SmartPaginate(paginate):
    def __init__(self, name='data',
                 use_prefix=False,
                 items_per_page=10,
                 max_items_per_page=0):
        super(SmartPaginate, self).__init__(name, use_prefix, items_per_page, max_items_per_page)


# -*- coding: utf-8 -*-
__author__ = 'vahid'

import sys
import os
import re
py_version = sys.version_info[:2]
try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages


install_requires = [
    'tgext.datahelpers >= 0.2.0',
    'sqlalchemy>=0.7.0',
    "tw2.core",
    "tw2.forms",
    "tw2.tinymce",
    "tw2.recaptcha",
]

# reading pymlconf version (same way sqlalchemy does)
with open(os.path.join(os.path.dirname(__file__), 'tgtoolbox', '__init__.py')) as v_file:
        package_version = re.compile(r".*__version__\s*=\s*[\"'](.*?)[\"']", re.S).match(v_file.read()).group(1)


setup(
    name='tgtoolbox',
    version=package_version,
    description='',
    author='Vahid Mardani',
    author_email='vahid.mardani@gmail.com',
    packages=find_packages(exclude=['ez_setup']),
#    package_data={'tgtoolbox': ['widgets/templates/*/*',
#                                'smart_crud/templates/*/*']},
    install_requires=install_requires,
    include_package_data=True,
    zip_safe=True
)



# -*- coding: utf-8 -*-
__author__ = 'vahid'

import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.header import Header
from tg.render import render
import tg


class Mailer(object):
    def __init__(self):
        self.port = 465
        self.host = tg.config.get('mailing.smtp_server')
        self.username = tg.config.get('mailing.smtp_username')
        self.password = tg.config.get('mailing.smtp_password')
        self.sender = tg.config.get('mailing.from_address')

    def _get_headers(self, subject, to):
        return "\r\n".join(["From: " + self.sender,
                            "Subject: " + subject,
                            "To: " + to,
                            "MIME-Version: 1.0",
                            "Content-Type: text/html"])


    def send(self, to=None, subject=None, body=None):
        message = MIMEMultipart('alternative')
        message.set_charset('utf8')
        message['FROM'] = self.sender


        # This solved the problem with the encode on the subject. #100
        message['Subject'] = Header(
            subject,
            'UTF-8'
        ).encode()

        message['To'] = to

        # And this on the body
        _attach = MIMEText(body, 'html', 'UTF-8')
        message.attach(_attach)

        session = smtplib.SMTP_SSL(self.host, port=self.port)
        session.login(self.username, self.password)
        try:
            session.sendmail(self.sender, to, message.as_string())
        finally:
            session.close()


    def send_template(self, to=None, subject=None, template=None, **kwargs):
        self.send(to=to, subject=subject, body=render(kwargs, template_name=template))

        # headers = self._get_headers(subject, to)
        # session = smtplib.SMTP_SSL(self.host, port=self.port)
        # session.login(self.username, self.password)
        # try:
        #     session.sendmail(self.sender, to, headers + "\r\n\r\n" + body)
        # finally:
        #     session.close()
